/******************************************************************************
 * Copyright (c) 2012-2017 All Rights Reserved, http://www.evocortex.com      *
 *  Evocortex GmbH                                                            *
 *  Emilienstr. 1                                                             *
 *  90489 Nuremberg                                                            *
 *  Germany                                                                   *
 *                                                                            *
 * Contributors:                                                              *
 *  Initial version for Linux 64-Bit platform supported by Fraunhofer IPA,    *
 *  http://www.ipa.fraunhofer.de                                              *
 *****************************************************************************/

#pragma once
#pragma comment(lib, "strmiids")

#define SKIP_DXTRANS
//#define SHOW_DEBUG_RENDERER
#define _CRT_SECURE_NO_WARNINGS

/* If you are having the problem you can't open dxtrans.h:
 * Open qedit.h and add this to the start of the file:
 *
 * #ifdef SKIP_DXTRANS
 * #define __IDxtAlphaSetter_INTERFACE_DEFINED__
 * #define __IDxtJpeg_INTERFACE_DEFINED__
 * #define __IDxtKey_INTERFACE_DEFINED__
 * #define __IDxtCompositor_INTERFACE_DEFINED__
 * #endif
 *
 * Also replace the line 
 * #include "dxtrans.h"
 * with:
 * #ifndef SKIP_DXTRANS
 * #include "dxtrans.h"
 * #endif
 */

#include <windows.h>
#include <dshow.h>
#include <map>
#include "IRDevice.h"
#include "irdirectsdk_defs.h"

namespace evo
{

#ifndef MAXLONGLONG
#define MAXLONGLONG 0x7FFFFFFFFFFFFFFF
#endif

#ifndef MAX_DEVICES
#define MAX_DEVICES 8
#endif

#ifndef MAX_DEVICE_NAME
#define MAX_DEVICE_NAME 80
#endif

#ifndef BITS_PER_PIXEL
#define BITS_PER_PIXEL 16
#endif

class IRDeviceDS;
struct ISampleGrabber;
class CallbackHandler;

/**
 * @class IRDeviceDS
 * @brief DirectShow device interface (Windows platforms)
 * @author Stefan May (Evocortex GmbH), Matthias Wiedemann (Optris GmbH)
 */
class __IRDIRECTSDK_API__ IRDeviceDS : public IRDevice
{
public:

  /**
   * Destructor
   */
  ~IRDeviceDS();

  /**
   * Factory method
   * @param[in] params device parameters
   */
  static IRDeviceDS*      createInstance(IRDeviceParams &params);

  /**
   * Start video streaming
   * @return success==true
   */
  int                     startStreaming();

  /**
   * Stop video streaming
   * @return success==true
   */
  int                     stopStreaming();

  /**
   * Get ID of device. If multiple devices are present, the ID is increased by 1 for each device. The fist device has the ID==1.
   * @return ID
   */
  int                     getId();

  /**
   * Get readable name of attached device
   * @return name of device type
   */
  const char*             getFriendlyName();  

  /**
   * Callback for getting access to raw data. It is called from a DirectShow thread.
   * @param[in] cb pointer to callback function
   */
  void                    setCallback(fptrIRRawFrame cb);

  /**
   * Set client class instance getting informed when new data is available
   * @param[in] client client instance
   */
  void                    setClient(IRImagerClient* client);

  /**
   * Run device, i.e., treat main thread in blocking mode. The data stream is not influence by this method.
   * It is implemented for the user's convenience, if the main thread is desired to be occupied.
   * User do not need to call this method, if they want to use their own idle loop.
   */
  virtual void            run();

  /**
   * Exit blocking run method, if previously called.
   */
  virtual void            exit();

  /**
  * Check if device was already opened
  * @return state (open / closed)
  */
  bool isOpen();

  /**
   *
   **/
  virtual IRDeviceError  getFrame(unsigned char* buffer, double* timestamp = NULL);

private:

  /**
  * Constructor
  */
  IRDeviceDS();

  static IRDeviceDS*      createVideoGraph(unsigned long serial, int videoFormatIndex);

  void                    initializeGraph();

  void                    initializeMediaControl();

  void                    releaseGraph();

  void                    onRawFrame(unsigned char* data, int len);

  static unsigned long    determineSerialNumber(wchar_t* deviceName);

  static void             deleteMediaType(AM_MEDIA_TYPE *pmt);

  static int              _id;

  char*                   _friendlyName;

  WCHAR*                  _filterName;

  IBaseFilter*            _sourceFilter;

  IBaseFilter*            _sampleGrabberFilter;

  IBaseFilter*            _nullRenderer;

  ISampleGrabber*         _sampleGrabber;

  ICaptureGraphBuilder2*  _builder;

  IMediaControl*          _control;

  IFilterGraph2*          _graph;

  CallbackHandler*        _callbackHandler;

  IMediaEventEx*          _event;

  HANDLE                  _wakeUp;

};

} // namespace
